/*
 * Copyright (C) 1996,1997,1998
 *	SHUDOH Kazuyuki
 *
 * Permission to use, copy, modify and distribute this software
 * and its documentation for non-commercial purposes and without
 * fee is hereby granted provided this copyright notice.
 *
 * $Id$
 */

package NET.shudo.moba.io;

import java.io.OutputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import NET.shudo.moba.reflect.*;
import NET.shudo.moba.lang.MobaThread;	// for method loadLibrary()


/**
 * Class externalizer<BR>
 * VM Υ饹򥯥饹եΥեޥåȤǽϤ롣
 *
 * @author SHUDOH Kazuyuki
 */
public final
class ClassWriter extends DataOutputStream {
  public ClassWriter(OutputStream out) {
    super(out);

    // native methods 饤֥Υ
    MobaThread.loadLibrary();
  }

  public ClassWriter() {
    this(System.out);
  }


  public void writeClass(Class clazz) throws IOException {
    // attrib. ̾ index
    int utfConstantValueIndex = ConstantPool.utf8Index(clazz, "ConstantValue");
    int utfCodeIndex = ConstantPool.utf8Index(clazz, "Code");
    int utfExceptionsIndex = ConstantPool.utf8Index(clazz, "Exceptions");
    int utfLineNoTblIndex = ConstantPool.utf8Index(clazz, "LineNumberTable");
    int utfLVarTblIndex = ConstantPool.utf8Index(clazz, "LocalVariableTable");
    int utfSourceFileIndex = ConstantPool.utf8Index(clazz, "SourceFile");
    int utfDeprecatedIndex = ConstantPool.utf8Index(clazz, "Deprecated");


    MetaClass mc = MetaClass.get(clazz);
    int i, j;

    // 饹γƼ
    int cpCount = mc.constantpoolCount();
    int intfCount = mc.interfaceCount();
    int fieldCount = mc.fieldCount();
    int methodCount = mc.methodCount();
    String sourceName = sourceName(mc.classClass());

    // 줾 attrib. ¸ߤ뤫ʤ (ѿ *Exists)
    boolean sourceFileExists =
	((sourceName != null) && (utfSourceFileIndex > 0));


    // magic number
    writeInt(0xCafeBabe);

    // major, minor version
    writeShort(3);	// minor
    writeShort(45);	// major

    // constantpool count
    writeShort(cpCount);

    // constantpool
    ConstantPool.writeInClassfile(this, clazz);

    // access flag
    writeShort(Access.maskInClassfile(mc.access()));

    // this class
    writeShort(ConstantPool.classIndex(clazz, clazz));

    // super class
    {
      MetaClass superMc = mc.superClass();
      if (superMc != null) {
	writeShort(ConstantPool.classIndex(clazz,
			mc.superClass().classClass()));
      }
      else {
	writeShort(0);
      }
    }

    // interface count
    writeShort(intfCount);

    // interface
    writeIFInClassfile(mc.classClass());

    // field count
    writeShort(fieldCount);


    // field
    for (i = 0; i < fieldCount; i++) {
      MetaField mf = null;
      int access = 0;
      try {
	mf = mc.field(i);
      }
      catch (FieldNotFoundException e) {
	// not reached.
      }
      access = mf.access();

      // access
      writeShort(Access.maskInClassfile(access));
      // name index
      writeShort(ConstantPool.utf8Index(clazz, mf.name()));
      // signature index
      writeShort(ConstantPool.utf8Index(clazz, mf.signature()));
      // attrib. count, attrib.
      if (Access.isStatic(access) &&
		(Access.isValknown(access) || Access.isFinal(access))) {
	if (utfConstantValueIndex >= 0) {
	  // attrib. count
	  writeShort(1);

	  // attrib. `ConstantValue'
	  writeShort(utfConstantValueIndex);	// `ConstantValue' in CP
	  writeInt(2);		// attrib. length
	  if (mf.isDouble()) {
	    writeShort(ConstantPool.doubleIndex(clazz,
					mf.getStaticDoubleValue()));
	  }
	  else if (mf.isLong()) {
	    writeShort(ConstantPool.longIndex(clazz, mf.getStaticLongValue()));
	  }
	  else if (mf.isFloat()) {
	    writeShort(ConstantPool.floatIndex(clazz, mf.getStaticFloatValue()));
	  }
	  else if (mf.isByte() || mf.isChar() || mf.isShort() || mf.isInt()) {
	    writeShort(ConstantPool.intIndex(clazz, mf.getStaticIntValue()));
	  }
	  else if (mf.isBoolean()) {
	    writeShort(ConstantPool.booleanIndex(clazz,
					mf.getStaticBooleanValue()));
	  }
	  else if (mf.isObject()) {
	    Object obj = mf.getStaticObjectValue();
	    if (obj instanceof java.lang.String) {
	      writeShort(ConstantPool.stringIndex(clazz, (String)obj));
	    }
	    else {
	      // not reached.
	      writeShort(0);
	    }
	  }
	  else {
	    // not reached.
	    writeShort(0);
	  }	// եɤηȽ
	}
	else
	  writeShort(0);
		// CP  (UTF8) `ConstantValue' Ĥ뤫
      }
      else
	writeShort(0);
		// (mod & ACC_STATIC) && (mod & ACC_VALKNOWN)
    }	// եɤé


    // method count
    writeShort(methodCount);


    // method
    for (i = 0; i < methodCount; i++) {
      MetaMethod mm = null;
      int access = 0;
      try {
	mm = mc.method(i);
      }
      catch (FieldNotFoundException e) {
	// not reached.
      }
      access = mm.access();

      // access
      writeShort(Access.maskInClassfile(access));
      // name index
      writeShort(ConstantPool.utf8Index(clazz, mm.name()));
      // signature index
      writeShort(ConstantPool.utf8Index(clazz, mm.signature()));
      // attrib. count, attrib.
      //	attribute `Deprecated' б
      {
	int methodAttrCount = 0;	// attrib. Code
	int thrownExLength = thrownExceptionLength(mm);
	boolean thrownExExists =
		((thrownExLength > 0) && (utfExceptionsIndex > 0));

	// attrib. count
	if (!Access.isNative(access))  methodAttrCount++;
	if (thrownExExists)  methodAttrCount++;
	writeShort(methodAttrCount);

	if (!Access.isNative(access)) {
	  // attrib. `Code'
	  {
	    writeShort(utfCodeIndex);	// `Code' in CP

	    long codeAttrLen;
	    long codeLength = codeLength(mm);
	    int catchFrameLength = catchFrameLength(mm);
	    int lineNoTableLength = lineNoTableLength(mm);
	    int localVarTableLength = localVarTableLength(mm);
	    boolean lineNoTableExists =
		((lineNoTableLength > 0) && (utfLineNoTblIndex > 0));
	    boolean localVarTableExists =
		((localVarTableLength > 0) && (utfLVarTblIndex > 0));

	    // attrib. length
	    codeAttrLen = 12 + codeLength + 8 * catchFrameLength;
	    if (lineNoTableExists) {
	      codeAttrLen += 8 + 4 * lineNoTableLength;
	    }
	    if (localVarTableExists) {
	      codeAttrLen += 8 + 10 * localVarTableLength;
	    }
	    writeInt((int)codeAttrLen);	// down cast

	    // max stack
	    writeShort(maxStack(mm));
	    // max locals
	    writeShort(maxLocals(mm));
	    // code length
	    writeInt((int)codeLength);	// down cast

	    // code
	    writeCode(mm, this);

	    // catch frame length
	    writeShort(catchFrameLength);
	    for (j = 0; j < catchFrameLength; j++) {
	      writeShort(catchFrameStartPC(mm, j));
	      writeShort(catchFrameEndPC(mm, j));
	      writeShort(catchFrameHandlerPC(mm, j));
	      writeShort(catchFrameExceptionIndexInCP(mm, j));
	    }

	    // attrib. count
	    int codeAttrCount = 0;
	    if (lineNoTableExists)  codeAttrCount++;
	    if (localVarTableExists)  codeAttrCount++;
	    writeShort(codeAttrCount);

	    // attrib. `LineNumberTable'
	    if (lineNoTableExists) {
	      writeShort(utfLineNoTblIndex);	// `LineNumberTable' in CP

	      // attrib. length
	      int attrLen = 2 + 4 * lineNoTableLength;
	      writeInt(attrLen);

	      writeShort(lineNoTableLength);

	      for (j = 0; j < lineNoTableLength; j++) {
		writeShort(lineNoTablePC(mm, j));
		writeShort(lineNoTableLineNo(mm, j));
	      }
	    }

	    // attrib. `LocalVariableTable'
	    if (localVarTableExists) {
	      writeShort(utfLVarTblIndex);	// `LovalVariableTable' in CP

	      // attrib .length
	      int attrLen = 2 + 10 * localVarTableLength;
	      writeInt(attrLen);

	      writeShort(localVarTableLength);

	      for (j = 0; j < localVarTableLength; j++) {
		writeShort(localVarTablePC(mm, j));
		writeShort(localVarTableLife(mm, j));
		writeShort(localVarTableNameIndex(mm, j));
		writeShort(localVarTableSigIndex(mm, j));
		writeShort(localVarTableSlot(mm, j));
	      }
	    }
	  }	// attrib. `Code'
	}	// native method ɤ

	// attrib. `Exceptions'
	if (thrownExExists) {
	  writeShort(utfExceptionsIndex);	// `Exceptions' in CP

	  // attrib. length
	  int attrLen = 2 + 2 * thrownExLength;
	  writeInt(attrLen);

	  writeShort(thrownExLength);

	  for (j = 0; j < thrownExLength; j++) {
	    writeShort(thrownExceptionIndexInCP(mm, j));
	  }
	}
      }	// attrib. count, attrib.
    }	// ᥽åɤé


    // attrib. count
    int classAttribCount = 0;
    if (sourceFileExists)  classAttribCount++;
    writeShort(classAttribCount);

    // attrib. `Source'
    if (sourceFileExists) {
      writeShort(utfSourceFileIndex);	// `Code' in CP

      // attrib. length
      writeInt(2);

      writeShort(ConstantPool.utf8Index(clazz, sourceName));
    }


    this.flush();
  }


  // 饹˴ؤ᥽åɷ

  /**
   * classfile ΥեޥåȤ interface Ϥ롣
   */
  private native void writeIFInClassfile(Class clazz)
	throws IOException;


  /**
   * ե֤̾
   */
  public static native String sourceName(Class clazz);


  /**
   * ClassWriter 륯饹եĹ֤
   */
  public long classfileLength(Class clazz) {
    long classfileLen = 0;

    // attrib. ̾ index
    int utfConstantValueIndex =
	ConstantPool.utf8Index(clazz, "ConstantValue");
    int utfCodeIndex = ConstantPool.utf8Index(clazz, "Code");
    int utfExceptionsIndex = ConstantPool.utf8Index(clazz, "Exceptions");
    int utfLineNoTblIndex =
	ConstantPool.utf8Index(clazz, "LineNumberTable");
    int utfLVarTblIndex =
	ConstantPool.utf8Index(clazz, "LocalVariableTable");
    int utfSourceFileIndex = ConstantPool.utf8Index(clazz, "SourceFile");


    MetaClass mc = MetaClass.get(clazz);
    int i, j;

    // 饹γƼ
    int cpCount = mc.constantpoolCount();
    int intfCount = mc.interfaceCount();
    int fieldCount = mc.fieldCount();
    int methodCount = mc.methodCount();
    String sourceName = sourceName(clazz);

    // 줾 attrib. ¸ߤ뤫ʤ (ѿ *Exists)
    boolean sourceFileExists =
	((sourceName != null) && (utfSourceFileIndex > 0));


    if (classfileLen >= 0)  return classfileLen;


    // file Ƭ
    classfileLen = 10;
    classfileLen += ConstantPool.lengthInClassfile(clazz);
    classfileLen += 8;
    classfileLen += 2 * intfCount;


    // field count
    classfileLen += 2;
    // field
    for (i = 0; i < fieldCount; i++) {
      MetaField mf = null;
      int access = 0;
      try {
	mf = mc.field(i);
      }
      catch (FieldNotFoundException e) {
	// not reached.
      }
      access = mf.access();

      classfileLen += 8;
      if (Access.isStatic(access) && Access.isValknown(access)) {
	if (utfConstantValueIndex >= 0)
	  classfileLen += 8;
      }
    }	// եɤ


    // method count
    classfileLen += 2;
    // method
    for (i = 0; i < methodCount; i++) {
      MetaMethod mm = null;
      int access = 0;
      try {
	mm = mc.method(i);
      }
      catch (FieldNotFoundException e) {
	// not reached.
      }
      access = mm.access();

      classfileLen += 8;

      // attrib. `Code'
      if (!Access.isNative(access)) {
	classfileLen += 14;
	classfileLen += codeLength(mm);

	int catchFrameLength = catchFrameLength(mm);
	int lineNoTableLength = lineNoTableLength(mm);
	int localVarTableLength = localVarTableLength(mm);
	boolean lineNoTableExists =
		((lineNoTableLength > 0) && (utfLineNoTblIndex > 0));
	boolean localVarTableExists =
		((localVarTableLength > 0) && (utfLVarTblIndex > 0));

	// catch frame
	classfileLen += 2;
	classfileLen += 8 * catchFrameLength;

	// attrib. in attrib. `Code'
	classfileLen += 2;
	// attrib. `LineNumberTable'
	if (lineNoTableExists) {
	  classfileLen += 8 + 4 * lineNoTableLength;
	}
	// attrib. `LocalVariableTable'
	if (localVarTableExists) {
	  classfileLen += 8 + 10 * localVarTableLength;
	}
      }	// attrib. `Code'

      // attrib. `Exceptions'
      int thrownExLength = thrownExceptionLength(mm);
      boolean thrownExExists =
		((thrownExLength > 0) && (utfExceptionsIndex > 0));
      if (thrownExExists) {
	classfileLen += 8 + 2 * thrownExLength;
      }
    }	// ᥽åɤé

    classfileLen += 2;

    // attrib. `Source'
    if (sourceFileExists)  classfileLen += 8;

    return classfileLen;
  }


  // ᥽åɤ˴ؤ᥽åɷ

  public static native int maxStack(MetaMethod mm);
  public static native int maxLocals(MetaMethod mm);
  public static native long codeLength(MetaMethod mm);

  public static native int catchFrameLength(MetaMethod mm);
  public static native int catchFrameStartPC(MetaMethod mm, int CFIndex);
  public static native int catchFrameEndPC(MetaMethod mm, int CFIndex);
  public static native int catchFrameHandlerPC(MetaMethod mm, int CFIndex);
  public static native int catchFrameExceptionIndexInCP(MetaMethod mm, int CFIndex);

  public static native int lineNoTableLength(MetaMethod mm);
  public static native int lineNoTablePC(MetaMethod mm, int index);
  public static native int lineNoTableLineNo(MetaMethod mm, int index);

  public static native int localVarTableLength(MetaMethod mm);
  public static native int localVarTablePC(MetaMethod mm, int index);
  public static native int localVarTableLife(MetaMethod mm, int index);
  public static native int localVarTableNameIndex(MetaMethod mm, int index);
  public static native int localVarTableSigIndex(MetaMethod mm, int index);
  public static native int localVarTableSlot(MetaMethod mm, int index);

  public static native int thrownExceptionLength(MetaMethod mm);
  public static native int thrownExceptionIndexInCP(MetaMethod mm, int index);

  public static native void writeCode(MetaMethod mm, OutputStream out);
}
