/*
 * Copyright (C) 1996,1997,1998,1999
 *	SHUDO Kazuyuki
 *
 * Permission to use, copy, modify and distribute this software
 * and its documentation for non-commercial purposes and without
 * fee is hereby granted provided this copyright notice.
 *
 * $Id$
 */

package NET.shudo.moba.io;

import java.io.InputStream;
import java.io.DataInputStream;
import java.io.PrintStream;
import java.io.BufferedInputStream;
import java.io.IOException;
import java.util.Vector;
import java.util.Enumeration;
import java.util.Stack;
import NET.shudo.moba.reflect.*;


/**
 * Object de-serializer$B!#(B
 * java.io.ObjectInputStream $B$KAjEv$9$k!#(B
 *
 * @author SHUDO Kazuyuki
 */
public
class ObjectReader extends DataInputStream {
  /**
   * $B$9$G$K(B de-serialize $B$7$?(B object $B$NI=!#(B
   */
  private Vector objectArray = new Vector();

  private int nullCount = 0;

  /**
   * $B%/%i%9%m!<%@!#(B
   */
  private ClassLoader loader = null;


  public ObjectReader(InputStream in, ClassLoader loader) throws IOException {
    this(in);

    if (loader != null)  this.loader = loader;
  }
  public ObjectReader(InputStream in) throws IOException {
    super((in.markSupported()) ? in : (new BufferedInputStream(in)));

    if (!this.markSupported()) {
      // not reached.
      throw new IOException("failed to make markSupported stream");
    }

    this.loader = this.getClass().getClassLoader();
  }


  /**
   * $B;H$&$Y$-(B ClassLoader $B$rJV$9!#(B
   */
  public ClassLoader classLoader() {  return this.loader;  }


  /**
   * subclass$B$GFCDj$N%/%i%9$KBP$9$k=hM}$r$9$k$?$a$N(Bhook$B!#(B
   * $BFCDj%/%i%9$+$I$&$+H=Dj$9$k!#(B
   *
   * @return true if the specified class is the target class.
   */
  protected boolean isSpecificClass(Class clazz) {
    return false;
  }

  /**
   * subclass$B$GFCDj$N%/%i%9$KBP$9$k=hM}$r$9$k$?$a$N(Bhook$B!#(B<BR>
   * $B%$%s%9%?%s%9$N%a%s%P$r@_Dj$7!"JV$9!#(B<BR>
   * $BFCDj$N%/%i%9$H$=$N(B supeclass $B$K$D$$$F$N@UG$$,$"$k!#(B
   *
   * @param obj instantiate $B:Q$_$N%*%V%8%'%/%H(B
   * @return true if the specified class is the target class.
   */
  protected boolean readSpecificClass(Class clazz, Object obj)
	throws StreamFormatException
		,IOException, ClassNotFoundException
		,InstantiationException, IllegalAccessException {
    return false;
  }

  /**
   * subclass$B$GFCDj$N%/%i%9$KBP$9$k=hM}$r$9$k$?$a$N(Bhook$B!#(B<BR>
   * $B40@.$7$?%$%s%9%?%s%9$KBP$9$k=hM}$r9T$&!#(B
   *
   * @return true if the specified class is the target class.
   */
  protected boolean postReadSpecificClass(Object obj) {
    return false;
  }


  public Object readObject()
	throws StreamFormatException
		,IOException, ClassNotFoundException
		,InstantiationException, IllegalAccessException {
    Object obj = null;
    MetaClass metaClass = null;

    Stack metaClassStack = new Stack();
	/* instance var. $B$K$7$F$O$$$1$J$$(B */


    // null reference
    if (readNull()) {
      return null;
    }

    // $B%$%s%G%C%/%9(B
    obj = readIndex();
    if (obj != null) {
      return obj;
    }

    // $BFCJL07$$$9$k%/%i%9(B
    obj = checkSpecialClasses();
    if (obj != null) {
      objectArray.addElement(obj);

      return obj;
    }


    // $B%/%i%9L>$rFI$_9~$_(B
    String className = readUTF();

    metaClass = MetaClass.get(className, this.loader);

    if (metaClass.isArray()) {		// $BG[Ns(B
      int len = readInt();	// $BG[Ns$ND9$5(B
      if (len == -1)  return null;

      obj = (Object)metaClass.newInstance(len, this.loader);
	// $BG[Ns%*%V%8%'%/%H$r:n@.(B
	// $B$3$N(B ObjectReader $B$,(B load $B$5$l$?(B ClassLoader $B$+$i%/%i%9$r%m!<%I(B

      objectArray.addElement(obj);	// $BEPO?(B

      if (className.startsWith("[L") || className.startsWith("[[")) {
	// $B%*%V%8%'%/%H(B or $BG[Ns(B $B$NG[Ns(B
	fillObjectArray(obj, metaClass);
      }
      else {
	// $B4pK\7?$NG[Ns(B
	fillBaseTypeArray(obj, className);
      }

      return obj;
    }


    // $B%$%s%9%?%s%9$r:n$k(B
    obj = metaClass.newInstance();
    // $B:n$C$?%*%V%8%'%/%H$rEPO?(B
    objectArray.addElement(obj);

    // $B%/%i%9<+A0$N(B readExternal() $B$r5/F0(B
    if (metaClass.isStreamable()) {
      ((Streamable)obj).readFromStream(this);
      return obj;
    }

    // MetaClass $B$r%9%?%C%/$K@Q$`(B
    while (metaClass != null) {
      metaClassStack.push(metaClass);
      if (isSpecificClass(metaClass.classClass()))
	break;
      metaClass = metaClass.superClass();
    }

    // java.lang.Object $B$+$i(B subclass $B$rC)$k(B
    while (!(metaClassStack.empty())) {
      metaClass = (MetaClass)metaClassStack.pop();

      // ObjectReader $B$N(B subclass $B$GFCJL07$$$9$k%/%i%9(B
      if (readSpecificClass(metaClass.classClass(), obj))
	continue;

      // $B%/%i%9<+A0$N(B readObject() $B$r5/F0(B
      if (metaClass.hasOwnReadObject()) {
	invokeOwnObjectReader(obj);

	int term = readByte();
	if (((byte)term) != (byte)0xff) {
	  throw new StreamFormatException("Illegal stream format");
	}

	continue;
      }

      // read members
      for (int i = 0; i < metaClass.fieldCount(); i++) {// follow fields
	MetaField field = null;
	try {
	  field = metaClass.field(i);
	}
	catch (FieldNotFoundException e) {
	  // not reached
	  e.printStackTrace();  System.exit(1);
	}

	// a field is not to be serialized
	if (!field.isAccessible())  continue;

	// object or array
	else if (field.isObject()) {
	  field.setObjectValue(obj, readObject());
	}
	// primitive type
	else {
	  readBaseType(obj, field);
	}
      } // for (int i = 0; i < metaClass.fieldCount(); i++)  // follow fields
    }	// while (!(metaClassStack.empty()))  follow subclasses


    // terminater
    {
      int term = readByte();
      if ((byte)term != (byte)0xff) {
	throw new StreamFormatException(
		"Illegal stream format(can't find terminator)");
      }
    }


    // revise location dependent information
    if (reviseLocDep) {
      // java.io.FileDescriptor#fd
      if (obj instanceof java.io.FileDescriptor) {
//System.out.println("readObject(): FileDescriptor is found.");
//System.out.flush();
	MetaField fdField = null;
	try {
	  fdField = metaClass.field("fd");
	}
	catch (FieldNotFoundException e) {
	  // not reached.
	  System.err.println("a field java.io.FileDescriptor#fd is not found.");
	  e.printStackTrace();
	}
	if (fdField != null) {
	  int fd = fdField.getIntValue(obj);
	  if (fd > 2) {		// fd is not stdin, stdout or stderr
	    fdField.setIntValue(obj, -1);	// disable the fd
	  }
//System.out.println("readObject(): set FileDescriptor#fd to -1");
//System.out.flush();
	}
      }
    }


    // $BFCJL07$$$9$k%/%i%9$K$D$$$F$N8e;OKv(B
    postReadSpecificClass(obj);


    return obj;
  }


  /**
   * $B0LCV0MB8>pJs$rJd@0$9$k$+$I$&$+$N%U%i%0!#(B
   */
  private boolean reviseLocDep = true;

  /**
   * $B0LCV0MB8>pJs$rJd@0$9$k$+$I$&$+@_Dj$9$k!#(B
   */
  public boolean reviseLocationDependence(boolean flag) {
    boolean old = reviseLocDep;
    reviseLocDep = flag;
    return old;
  }


  /**
   * $B%/%i%9<+A0$N(B readObject() $B$r8F$V!#(B
   */
  private native void invokeOwnObjectReader(Object obj)  throws IOException;


  /**
   * demarshal $B:Q$_%*%V%8%'%/%H$N%F!<%V%k$NCf?H$rI=<($9$k!#(B(for debug)
   */
  public void showObjectTable(PrintStream out) {
    int i = 0;
    for (Enumeration e = objectArray.elements(); e.hasMoreElements(); ) {
      out.println(i++ + ": " + (e.nextElement()));
    }
  }


  /**
   * writeObject() by itself.
   */
  private void writeObject(ObjectWriter out) throws IOException {
  }
  /**
   * readObject() by itself.
   */
  private void readObject(ObjectReader in)
	throws IOException, ClassNotFoundException {
  }


  /*
   * input.
   */

  private boolean readNull() throws IOException {
    byte b1, b2;

    // nulls continues
    if (nullCount > 0) { nullCount--;  return true; }

    mark(2);

    b1 = readByte();  b2 = readByte();
    if (b1 == (byte)0xff) {
      if (b2 == (byte)0xff) {
	return true;
      }
      else if (b2 == (byte)0xfe) {
	nullCount = readShort();
	nullCount--;
	return true;
      }
    }

    reset();
    return false;
  }

  private Object readIndex() throws IOException {
    byte b1, b2;

    mark(2);
    b1 = readByte();  b2 = readByte();
    if ((b1 == (byte)0xff) && (b2 == (byte)0xfd)) {
      int index = readInt();
//System.out.print("registered object[" + index + "]: ");  System.out.flush();
//System.out.println(objectArray.elementAt(index));  System.out.flush();

      return objectArray.elementAt(index);
    }

    reset();
    return null;
  }

  /**
   * handle a class should be processed specially.
   */
  private Object checkSpecialClasses() throws IOException {
    Object obj = null;

    // java.lang.String
    if ((obj = readString()) != null) {
      return obj;
    }
    // java.lang.Class
    else if ((obj = readClass()) != null) {
      return obj;
    }

    return null;
  }

  private String readString() throws IOException {
    byte b1, b2;

    mark(2);
    b1 = readByte();  b2 = readByte();
    if ((b1 == (byte)0xff) && (b2 == (byte)0xfc)) {
      String str = readUTF();
      str = str.intern();	// carefully, and not needed
      return str;
    }

    reset();
    return null;
  }

  private Class readClass() throws IOException {
    byte b1, b2;

    mark(2);
    b1 = readByte();  b2 = readByte();
    if ((b1 == (byte)0xff) && (b2 == (byte)0xfb)) {
      String className;
      Class clazz;

      className = readUTF();
      try {
	clazz = this.loader.loadClass(className);
      }
      catch (ClassNotFoundException e) {
	throw new IOException("loading a class " + className + " failed.");
      }
      return clazz;
    }

    reset();
    return null;
  }


  /**
   * read a primitive type value.
   */
  private final void readBaseType(Object obj, MetaField field)
	throws IOException {
    String sig = field.signature();
    char c = sig.charAt(0);

    switch(c) {
    case 'Z':
      field.setBooleanValue(obj, readBoolean());
      break;
    case 'B':
      field.setByteValue(obj, readByte());
      break;
    case 'C':
      field.setCharValue(obj, readChar());
      break;
    case 'I':
      field.setIntValue(obj, readInt());
      break;
    case 'J':
      field.setLongValue(obj, readLong());
      break;
    case 'F':
      field.setFloatValue(obj, readFloat());
      break;
    case 'D':
      field.setDoubleValue(obj, readDouble());
      break;
    default:
      // not reached
      // error handling is not implemented yet.
      break;
    }
  }

  /**
   * read an array of objects or arrays.
   */
  private final void fillObjectArray(Object obj, MetaClass mc)
	throws StreamFormatException
		,IOException, ClassNotFoundException
		,InstantiationException, IllegalAccessException {
    if (obj == null)  return;

    Object[] objArray = (Object[]) obj;

    for (int j = 0; j < objArray.length; j++) {
      objArray[j] = readObject();
    }
  }

  /**
   * read an array of primitive type values.
   */
  private final native void fillBaseTypeArray(Object array, String sig)
	throws IOException;
}
