/*
 * Copyright (C) 1996,1997,1998
 *	SHUDOH Kazuyuki
 *
 * Permission to use, copy, modify and distribute this software
 * and its documentation for non-commercial purposes and without
 * fee is hereby granted provided this copyright notice.
 *
 * $Id$
 */

#include <stdio.h>
#include <stdlib.h>	/* for malloc(), realloc(), free() */
#include <string.h>	/* for strncat() */
#include <netinet/in.h>	/* for [hn]to[nh][ls]() */

#include "buf.h"


/*
 * Local Functions
 */
static char *extend_capacity();
static char *ensure_capacity(size_t req_size);

/*
 * Variables
 */
static char *strbuf = NULL;	/* ХåեƬɥ쥹 */
static int write_size = 0;	/* ݤƤΰ礭 */
static char *write_ptr;		/* ˽񤭹٤ɥ쥹 */

static char *read_ptr;


/*
 * ExecEnv ǼХåեν
 */
char *write_init() {
#ifdef WRITE_VERBOSE
  printf("write_init() called.\n");
#endif
  if (!strbuf) {
    write_size = BUF_CHUNK_SIZE;
    strbuf = (char *)malloc(write_size);
  }
  write_ptr = strbuf;
#ifdef WRITE_VERBOSE
  printf("strbuf: 0x%x.\n", strbuf);
  fflush(stdout);
#endif

  return write_ptr;
}


/*
 * Хåե˥ԡ
 */
char *write_cat(char *src, int n) {
  ensure_capacity(n);

#ifdef WRITE_VERBOSE
  printf("write_cat(0x%x, %d) is called.\n", src, n);
  printf("write_ptr: 0x%x\n", (int)write_ptr);
  fflush(stdout);
#endif

  memcpy(write_ptr, src, n);
  write_ptr += n;

  return write_ptr;
}


char *write_string(char *src) {
  int len = strlen(src);
  unsigned short lenshort;

#ifdef WRITE_VERBOSE
  printf("write_string: `%s'(%d)\n", src, len);
  fflush(stdout);
#endif
  lenshort = (unsigned short)len;
  lenshort = htons(lenshort);
  write_cat((char *)&lenshort, 2);

  return write_cat(src, (int)len);
}


char *write_hton32(char *p) {
  unsigned long l = *((long *)p);
#ifdef WRITE_VERBOSE
  printf("write_htons32: %d\n", (int)l);  fflush(stdout);
#endif
  l = htonl(l);
  return write_cat((char *)&l, 4);
}


char *write_hton16(char *p) {
  unsigned short s = *((short *)p);
#ifdef WRITE_VERBOSE
  printf("write_htons16: %d\n", (int)s);  fflush(stdout);
#endif
  s = htons(s);
  return write_cat((char *)&s, 2);
}


char *write_hton64(char *p) {
  unsigned long long ll = *((unsigned long long *)p);
  unsigned long high;
  unsigned long low;

  high = (unsigned long)((ll >> 32) & 0xffffffff);
  low = (unsigned long)(ll & 0xffffffff);

  high = htonl(high);
  low = htonl(low);

  write_cat((char *)&high, 4);
  return write_cat((char *)&low, 4);
}


char *write_char(int c) {
  char data = (char)c;
  return write_cat(&data, 1);
}


/*
 * ХåեȤ
 */
char *write_content() {
  *write_ptr = '\0';
  return strbuf;
}

int write_length() {
  return (int)(write_ptr - strbuf);
}

/*
 * 񤭽ФѥХåեΥ
 */
int write_buf_length() {
  return write_size;
}


/*
 * ɤߤΤν
 */
extern void read_init(char *src) {
  read_ptr = src;
}


extern void read_string(char *dest) {
  unsigned short len;

  read_ntoh16((void *)&len);

  memcpy(dest, read_ptr, (size_t)len);

  read_ptr += len;
}


extern char *read_strdup() {
  char *ret;
  unsigned short len;

  read_ntoh16((void *)&len);

  if (!(ret = (char *)malloc(len + 1))) {
    return NULL;
  }
  memcpy(ret, read_ptr, (size_t)len);
  read_ptr += len;

  *(ret + len) = '\0';		/* terminate */
  return ret;
}


extern void read_ntoh32(char *dest) {
  unsigned long l;

  memcpy(&l, read_ptr, 4);
  l = ntohl(l);
  memcpy(dest, &l, 4);

  read_ptr += 4;
}


extern void read_ntoh16(char *dest) {
  unsigned short s;

  memcpy(&s, read_ptr, 2);
  s = ntohs(s);
  memcpy(dest, &s, 2);

  read_ptr += 2;
}


extern void read_char(char *dest) {
  memcpy(dest, read_ptr, 1);
  read_ptr++;
}


/*
 * Local Functions
 */

/*
 * ХåեΥĥ
 */
static char *extend_capacity() {
  if (!strbuf) {
    write_size = BUF_CHUNK_SIZE;
    strbuf = (char *)malloc(write_size);
    write_ptr = strbuf;
  }
  else {
    int write_ptr_offset = write_ptr - strbuf;

    write_size += BUF_CHUNK_SIZE;
    strbuf = (char *)realloc(strbuf, write_size);

    write_ptr = strbuf + write_ptr_offset;
  }

  return write_ptr;
}


/*
 * Хåեζμ¤˳
 * ݤ»ͤ NULl ֤
 */
static char *ensure_capacity(size_t req_size) {
  if (!strbuf) {
    write_size = ((req_size / BUF_CHUNK_SIZE) + 1) * BUF_CHUNK_SIZE;
    strbuf = (char *)malloc(write_size);
    write_ptr = strbuf;
  }
  else {
    int remain = write_size - (write_ptr - strbuf);
    if (remain < req_size) {
      int write_ptr_offset = write_ptr - strbuf;

#ifdef WRITE_VERBOSE
      printf("extend capacity.\n");
      printf("strbuf, write_ptr: 0x%x, 0x%x.\n", strbuf, write_ptr);
      printf("write_ptr_offset: %d.\n", write_ptr_offset);
      fflush(stdout);
#endif

      {
	char *orig_strbuf = strbuf;
	int orig_write_size = write_size;
	write_size +=
		(((req_size - remain) / BUF_CHUNK_SIZE) + 1) * BUF_CHUNK_SIZE;
#ifdef WRITE_VERBOSE
	printf("write_size: %d -> %d.", orig_write_size, write_size);
	fflush(stdout);
#endif
#ifdef HAVE_REALLOC
#ifdef WRITE_VERBOSE
	printf("REALLOC(0x%x, %d).\n", (int)strbuf, write_size);
	fflush(stdout);
#endif
	strbuf = (char *)realloc((void *)strbuf, write_size);
#else
 printf("MALLOC(%d).\n", write_size);  fflush(stdout);
	strbuf = (char *)malloc(write_size);
 printf("MEMCPY(0x%x, 0x%x, %d).\n", (int)strbuf, (int)orig_strbuf, orig_write_size);  fflush(stdout);
	memcpy(strbuf, orig_strbuf, orig_write_size);
 printf("FREE(0x%x).\n", orig_strbuf);  fflush(stdout);
	free(orig_strbuf);
#endif /* HAVE_REALLOC */
#ifdef WRITE_VERBOSE
	printf("REALLOC() returns 0x%x.\n", (int)strbuf);
	fflush(stdout);
#endif
	if (!strbuf) {
	  return NULL;
	}

	write_ptr = strbuf + write_ptr_offset;

#ifdef WRITE_VERBOSE
	printf("strbuf, write_ptr: 0x%x, 0x%x.\n", strbuf, write_ptr);
	printf("extend capacity done.\n");
	fflush(stdout);
#endif
      }
    }	/* if (remain < req_size) */
  }	/* if (!strbuf) */

  return write_ptr;
}
