/*
 * Copyright (C) 1996,1997,1998,1999
 *	SHUDO Kazuyuki
 *
 * Permission to use, copy, modify and distribute this software
 * and its documentation for non-commercial purposes and without
 * fee is hereby granted provided this copyright notice.
 *
 * $Id$
 */

package NET.shudo.moba.place;


import java.io.InputStream;
import java.io.OutputStream;
import java.io.FileInputStream;	// for optimization of throughput
import java.io.BufferedInputStream;	// for optimization of throughput
import java.io.IOException;
import java.net.Socket;
import java.net.InetAddress;
import NET.shudo.moba.lang.MobaThread;
import NET.shudo.moba.lang.PlaceAddress;
import NET.shudo.moba.io.ThreadReader;
import NET.shudo.moba.io.ObjectReader;


/**
 * $B$R$H$D$N%9%l%C%I$r<u$1F~$l!"5/F0$9$k%/%i%9!#(B<BR>
 * ThreadServer $B$+$i5/F0$5$l$k!#(B
 *
 * @see NET.shudo.moba.place.ThreadServer
 */
public
class ThreadImporter implements Runnable {
  InputStream in;
  OutputStream out;	// optional
  Socket sock;		// optional
  String filename;	// optional
  private boolean debug = false;


  protected ThreadImporter(Socket sock, boolean debug)
	throws IOException {
    this(sock.getInputStream(), debug);
    this.sock = sock;
  }
  protected ThreadImporter(String fname, boolean debug)
	throws IOException {
    this(new FileInputStream(fname), debug);
    this.filename = fname;
  }
  private ThreadImporter(InputStream in, boolean debug) {
    this.in = new BufferedInputStream(in, 1024 * 32);
//    this.in = in;
    this.debug = debug;
  }


  public void run() {
    boolean success = true;	// $B<u$1F~$l$N@.H](B

    // $B%/%i%9%m!<%@$rF@$k(B
    MobaClassLoader loader = null;
    {
      PlaceAddress classSrc = null;
      try {
	if (sock != null)	{	// ThreadExporter $B$H$d$j$H$j(B
	  int len, read;
	  len = in.read() << 8;  len |= in.read();
	  byte[] bary = new byte[len];
	  read = 0;
	  while (len > 0) {
	    read += in.read(bary, read, len);
	    len -= read;
	  }
	  String genedPlace = new String(bary);
if (debug) {
  System.out.println("ThreadImporter#run(): place on which the thread generated is `" + genedPlace + "'");
  System.out.flush();
}
	  classSrc = new PlaceAddress(genedPlace);
	}
	else if (filename != null)
	  classSrc = ThreadReader.generatedPlaceFromFile(filename);
      }
      catch (IOException e) {  if (debug)  e.printStackTrace();  }

      if (classSrc != null) {
	try {
	  loader = MobaClassLoader.get(classSrc);
	  loader.debug(debug);

	  // $B;2>H%+%&%s%H$rA}$d$9(B
	  loader.register();
	}
	catch (IOException e) {
	  success = false;
	  if (debug) { e.printStackTrace(); System.out.flush(); }
	}
      }
    }


    ObjectReader thrIn = null;
    try {
      thrIn = new ThreadReader(MobaPlace.threadTank, in, loader);
    }
    catch (Exception e) {
      success = false;
      if (debug) { e.printStackTrace(); System.out.flush(); }
    }



    // $B%9%l%C%I$N<u$1F~$l(B
    MobaThread thread = null;

    if (success) {
      try {
	if (debug)
	  System.out.println("ThreadImporter: now reading a thread..");
	thread = (MobaThread)thrIn.readObject();
      }
      catch (Exception e) {
	success = false;
	if (debug)  e.printStackTrace();
      }
      Logger.log(this, "receipt " + (success?"success":"failed") + ".");
      if (debug)
	System.out.println("ThreadImporter: reading is done (" +
		(success?"success":"failure") + ").");
    }

    if (success) {	// $B<u$1F~$l@.8y(B
      // notify
      PlaceEventNotifier notifier = MobaPlace.eventNotifier();
      if (notifier != null) {
	if (this.sock != null) {
	  notifier.comeNotify(
		sock.getInetAddress().getHostAddress(), -1, thread);
	}
	else if (this.filename != null) {
	  notifier.startedNotify(thread);
	}
      }
    }

    // close
    try {
      in.close();
      if (sock != null)  sock.close();
    }
    catch (IOException e) {  if (debug)  e.printStackTrace();  }

    if (success) {	// $B<u$1F~$l@.8y(B
      // $B<u$1F~$l$?%9%l%C%I$N(B resume
if (debug) {
  System.out.println("now resume the read thread.");
  System.out.flush();
}
      thread.resume();
if (debug) {
  System.out.println("listing by ThreadImporter (resumed thread):");
  (new NET.shudo.moba.util.ThreadLister()).listAllThreads(System.out);
  System.out.flush();
}

      // $B<u$1F~$l%9%l%C%I$N=*N;BT$A(B
      boolean joinDone;
      do {
	joinDone = true;
	try {
	  thread.join();
	}
	catch (InterruptedException e) { joinDone = false; }
      } while (!joinDone);

if (debug) {
System.out.println("listing by ThreadImporter (join done):");
(new NET.shudo.moba.util.ThreadLister()).listAllThreads(System.out);
System.out.flush();
}
    }	// if (success)
    else {
if (debug) {
System.out.println("Migration failed.");
System.out.flush();
}
    }


    // $B%/%i%9%m!<%@$N;2>H%+%&%s%H$r8:$i$9(B
    if (loader != null)  loader.unregister();
  }
}
