/**
  Sample code for object serializer
  in package NET.shudo.moba.
  Copy this file to the directory which is in CLASSPATH and compile:
    % javac SerializeObjectTest.java
  and execute:
    % java SerializeObjectTest

  @author SHUDOH Kazuyuki
 */

import java.io.InputStream;
import java.io.OutputStream;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;

import NET.shudo.moba.io.ObjectWriter;
import NET.shudo.moba.io.ObjectReader;
import NET.shudo.moba.io.Streamable;


/**
 * class ObjectWriter, ObjectReader Υƥȡ
 */
public class SerializeObjectTest {
  private static final String OBJ_FILE = "output";

  private void run() {
    // output
    try {
      OutputStream fout = new FileOutputStream(OBJ_FILE);
      ObjectWriter sout = new ObjectWriter(fout);

      LinkedString msg0 = new LinkedString("HelloWorld");
      LinkedString msg1 = new LinkedString("I'm a child");
      LinkedString msg2 = new LinkedString("I'm a grandchild");
      msg1.addChild(msg2);
      msg0.addChild(msg1);

      System.out.println(msg0);
      System.out.println(msg0.getChild());
      System.out.println(msg0.getChild().getChild());

      System.out.println("Writing to file \"" + OBJ_FILE + "\".");
      sout.writeObject(msg0);

      sout.flush();
      sout.close();
    }
    catch (Exception e) {
      System.err.println(e); e.printStackTrace(); System.exit(1);
    }

    // input
    try {
      InputStream fin = new FileInputStream(OBJ_FILE);
      ObjectReader sin = new ObjectReader(fin);

      System.out.println("Reading from file \"" + OBJ_FILE + "\".");
      LinkedString msg0 = (LinkedString)sin.readObject();

      sin.close();

      System.out.println(msg0);
      System.out.println(msg0.getChild());
      System.out.println(msg0.getChild().getChild());
    }
    catch (Exception e) {
      System.err.println(e); e.printStackTrace(); System.exit(1);
    }

    System.out.println("Succeed!");
  }

  public static void main(String argv[]) {
    new SerializeObjectTest().run();
  }
}


// wrapper class for String
class LinkedString {
//implements Streamable {
  private String str = "get it up!";
  LinkedString child = null;

  // constructor
  LinkedString() {}
  LinkedString(String initstr) {
    this.val(initstr);
  }

  public String toString() {
    return str;
  }

  // methods
  public String val() {
    return str;
  }

  public String val(String newstr) {
    str = new String(newstr);
    return str;
  }

  public void addChild(LinkedString newchild) {
    child = newchild;
  }

  public LinkedString getChild() {
    return child;
  }

/*
  public void writeToStream(ObjectWriter out) throws IOException {
    out.writeUTF(val());
    if (child != null) {
      out.write(0xff);
      child.writeToStream(out);
    }
    else {
      out.write(0x00);
    }
  }

  public void readFromStream(ObjectReader in) throws IOException {
    val(in.readUTF());

    int next = in.readByte();
    if (next != 0) {
      child = new LinkedString();
      child.readFromStream(in);
    }
  }
*/
}
